// Copyright 2019 Lawrence Livermore National Security, LLC and other
// Devil Ray Developers. See the top-level COPYRIGHT file for details.
//
// SPDX-License-Identifier: (BSD-3-Clause)


#include "gtest/gtest.h"

#include "t_utils.hpp"
#include "t_config.hpp"

#include <dray/io/blueprint_reader.hpp>

#include <dray/filters/reflect.hpp>
#include <dray/filters/mesh_boundary.hpp>
#include <dray/rendering/surface.hpp>
#include <dray/rendering/renderer.hpp>
#include <dray/array_registry.hpp>

#include <dray/utils/appstats.hpp>

#include <dray/math.hpp>

#include <fstream>
#include <stdlib.h>
//---------------------------------------------------------------------------//
bool
mfem_enabled()
{
#ifdef DRAY_MFEM_ENABLED
    return true;
#else
    return false;
#endif
}

TEST (dray_reflect, dray_reflect_2d)
{
  if(!mfem_enabled())
  {
    std::cout << "mfem disabled: skipping test that requires high order input " << std::endl;
    return;
  }
  // between tests, we should always start with 0 arrays, 0 usage
  EXPECT_EQ(dray::ArrayRegistry::number_of_arrays(),0);
  EXPECT_EQ(dray::ArrayRegistry::host_usage(),0);
  EXPECT_EQ(dray::ArrayRegistry::device_usage(),0);
  dray::ArrayRegistry::summary();
  dray::stats::StatStore::clear();
  if(dray::stats::StatStore::stats_supported())
  {
    dray::stats::StatStore::enable_stats();
  }

  std::string root_file = std::string (ASCENT_T_DATA_DIR) + "taylor_green_2d.cycle_000050.root";
  std::string output_path = prepare_output_dir ();
  std::string output_file =
  conduit::utils::join_file_path (output_path, "tg_2d_reflect");
  remove_test_image (output_file);

  dray::Collection collection = dray::BlueprintReader::load (root_file);

  dray::Vec<float,3> point = {0.f, 0.f, 0.f};
  dray::Vec<float,3> normal = {0.f, 1.f, 0.f};

  dray::Reflect reflector;
  reflector.plane(point, normal);
  dray::Collection reflected = reflector.execute(collection);

  dray::AABB<3> bounds;
  bounds.include(collection.bounds());
  bounds.include(reflected.bounds());

  dray::ColorTable color_table ("Spectral");

  // Camera
  const int c_width = 512;
  const int c_height = 512;

  std::string field_name = "density";

  dray::Camera camera;
  camera.set_width (c_width);
  camera.set_height (c_height);
  camera.reset_to_bounds (bounds);

  std::shared_ptr<dray::Surface> surface
    = std::make_shared<dray::Surface>(collection);
  surface->field(field_name);
  surface->color_map().color_table(color_table);
  surface->draw_mesh (true);
  surface->line_thickness(.1);

  std::shared_ptr<dray::Surface> surface2
    = std::make_shared<dray::Surface>(reflected);
  surface2->field(field_name);
  surface2->color_map().color_table(color_table);
  surface2->draw_mesh (true);
  surface2->line_thickness(.1);

  dray::Renderer renderer;
  renderer.add(surface);
  renderer.add(surface2);
  dray::Framebuffer fb = renderer.render(camera);

  fb.save(output_file);
  // note: dray diff tolerance was 0.2f prior to import
  EXPECT_TRUE (check_test_image (output_file,dray_baselines_dir(),0.05));
  fb.save_depth (output_file + "_depth");
  dray::stats::StatStore::write_ray_stats (output_file + "_stats",
                                           c_width, c_height);
}
